@@ -1,5 +1,7 @@
 #include <Python.h>
 
+#include <tools.h>
+
 #include <apply.h>
 #include <validate.h>
 // https://edcjones.tripod.com/refcount.html
@@ -9,14 +11,14 @@
 PyDoc_STRVAR(
     __doc__,
     "\n"
-    "apply(callable, *objects, _safe=True, _star=True, _finalizer=None, **kwargs)\n"
+    "apply(callable, *objects, _safe=True, _star=True, _finalizer=None, _strict=True, **kwargs)\n"
     "\n"
     "Compute the function using the leaf data of the nested objects as arguments.\n"
     "\n"
     "A `nested object` is either a python object (object, str, numpy array, torch\n"
-    "tensor, etc.) or a subclass of one of python's built-in containers (dict,\n"
-    "list, or tuple), that consists of other nested objects. The `leaf data`\n"
-    "is any non-contaier python object at the bottom of the nested structure.\n"
+    "tensor, etc.) or one of python's built-in containers (dict, list, or tuple),\n"
+    "that consists of other nested objects. The `leaf data` is any non-container\n"
+    "python object at the bottom of the nested structure.\n"
     "\n"
     "Parameters\n"
     "----------\n"
@@ -51,6 +53,13 @@ PyDoc_STRVAR(
     "\n"
     "    OMIT the `_finalizer` kwarg if finalization is NOT REQUIRED.\n"
     "\n"
+    "_strict : bool, default=True\n"
+    "    Whether to treat the subtypes of built-in containers as non-leaf nested\n"
+    "    containers and descend into them. NOTE, that when being rebuilt, subtypes\n"
+    "    REGRESS to their built-in base types.\n"
+    "\n"
+    "    NOTE `_strict` does not affect treatment of namedtuples (SEE caveat).\n"
+    "\n"
     "**kwargs : variable keyword arguments\n"
     "   The optional keyword arguments passed AS IS to the `callable` every\n"
     "   time it is invoked on the leaf data.\n"
@@ -78,6 +87,17 @@ PyDoc_STRVAR(
     "          being runtime-constructed sub-classes of tuples. Hence for them\n"
     "          only the order matters and not their fields' names.\n"
     "\n"
+    "Caveat on namedtuples\n"
+    "---------------------\n"
+    "`apply` treats namedtuples as nested containers regardless of the `_strict`\n"
+    "flag. This was designed intentionally, since NTs are tuples with attributes\n"
+    "identifying the items within, and as such can be viewed as immutable dict-like\n"
+    "structures. Detection of NTs, however is somewhat duck-typed and non-robust:\n"
+    "we check if a container's type ONLY from a `tuple` AND the object itself has\n"
+    "`_fields` attribute as suggested in this discussion:\n"
+    "\n"
+    "    https://mail.python.org/pipermail//python-ideas/2014-January/024886.html\n"
+    "\n"
     "Details\n"
     "-------\n"
     "For a single container `apply` with `_star=True` and a specified\n"
@@ -98,68 +118,6 @@ PyDoc_STRVAR(
 );
 
 
-PyObject *PyObject_CallWithSingleArg(
-    PyObject *callable,
-    PyObject *arg,
-    PyObject *kwargs)
-{
-    // much like `PyObject_CallOneArg`, but with optional kwargs:
-    //   create a one-element tuple, then call with it and kwargs
-    PyObject *single = PyTuple_New(1);
-    if(single == NULL) return NULL;
-
-    // `PyTuple_SET_ITEM` steals reference to `arg`, but we borrowed it from
-    //  the caller! Thus we first become its independent owner.
-    Py_INCREF(arg);
-    PyTuple_SET_ITEM(single, 0, arg);
-
-    // the called object increfs its returned value and transfers the ownership
-    //  to the caller, i.e. the act of RETURNING is in itself another reference
-    //  see https://docs.python.org/3/extending/extending.html#ownership-rules
-    //      https://stackoverflow.com/questions/57661466/
-    // (we may check refcounts of `output` and `arg` for `lambda x: x`)
-    PyObject *output = PyObject_Call(callable, single, kwargs);
-
-    // No need to relinquish ownership of `arg` (decref), since decrefing
-    //  a tuple decrefs all non-NULL items (tuple steals/assumes ownership,
-    //  instead of borrowing).
-    Py_DECREF(single);
-
-    return output;
-}
-
-
-PyObject *PyDict_SplitItemStrings(
-    PyObject *dict,
-    char *keys[],
-    const bool pop=false)
-{
-    // Pop the specified keys in a NULL-terminated str key list `keys` from
-    //  `dict` and put them into a new dict
-    // XXX warning, this manipulates the `dict` inplace unless `pop=false`!
-    PyObject* subdict = PyDict_New();
-    if (subdict == NULL)
-        return NULL;
-
-    // transfer (unless pop=false) objects' ownership from one dict into another
-    for(int p = 0; keys[p] != NULL; p++) {
-        PyObject* item = PyDict_GetItemString(dict, keys[p]);
-        if (item == NULL)
-            continue;
-
-        // PyDict_SetItem uses `Py_INCREF() to become an independent owner`
-        //  see https://docs.python.org/3/extending/extending.html#ownership-rules
-        PyDict_SetItemString(subdict, keys[p], item);  // increfs `item`
-
-        if(pop)
-            PyDict_DelItemString(dict, keys[p]);  // decrefs `item`
-    }
-
-    // could be an empty dict
-    return subdict;
-}
-
-
 PyObject* _apply(
     PyObject *callable,
     PyObject *main,
@@ -167,7 +125,8 @@ PyObject* _apply(
     const bool safe,
     const bool star,
     PyObject *kwargs,
-    PyObject *finalizer);
+    PyObject *finalizer,
+    const bool strict);
 
 
 static PyObject* _apply_dict(
@@ -177,7 +136,8 @@ static PyObject* _apply_dict(
     const bool safe,
     const bool star,
     PyObject *kwargs,
-    PyObject *finalizer)
+    PyObject *finalizer,
+    const bool strict)
 {
     Py_ssize_t len = PyTuple_GET_SIZE(rest);
     PyObject *key, *main_, *item_, *rest_ = PyTuple_New(len);
@@ -212,7 +172,7 @@ static PyObject* _apply_dict(
         }
 
         // `result` is a new object, for which we are now responsible
-        result = _apply(callable, main_, rest_, safe, star, kwargs, finalizer);
+        result = _apply(callable, main_, rest_, safe, star, kwargs, finalizer, strict);
         if(result == NULL) {
             Py_DECREF(rest_);
 
@@ -231,7 +191,8 @@ static PyObject* _apply_dict(
         Py_DECREF(result);
     }
 
-    // decrefing a tuple also decrefs all its items
+    // decrefing a tuple with only one reference, as is the case here, also
+    //  decrefs all its items.
     Py_DECREF(rest_);
 
     return output;
@@ -245,7 +206,8 @@ static PyObject* _apply_tuple(
     const bool safe,
     const bool star,
     PyObject *kwargs,
-    PyObject *finalizer)
+    PyObject *finalizer,
+    const bool strict)
 {
     Py_ssize_t len = PyTuple_GET_SIZE(rest);
     PyObject *main_, *item_, *rest_ = PyTuple_New(len);
@@ -272,7 +234,7 @@ static PyObject* _apply_tuple(
             PyTuple_SET_ITEM(rest_, j, item_);
         }
 
-        result = _apply(callable, main_, rest_, safe, star, kwargs, finalizer);
+        result = _apply(callable, main_, rest_, safe, star, kwargs, finalizer, strict);
         if(result == NULL) {
             Py_DECREF(rest_);
             Py_DECREF(output);
@@ -290,15 +252,12 @@ static PyObject* _apply_tuple(
     if(PyTuple_CheckExact(main))
         return output;
 
-    // Preserve namedtuple, devolve others to builtin tuples
-    // "isinstance(o, tuple) and hasattr(o, '_fields')" is the corect way.
-    //   https://mail.python.org/pipermail//python-ideas/2014-January/024886.html
-    //   https://bugs.python.org/issue7796
-    if(!PyObject_HasAttrString(main, "_fields"))
+    // Preserve namedtuple, devolve others to built-in tuples
+    if(!PyNamedTuple_CheckExact(main))
         return output;
 
-    // since `namedtuple`-s are immutable and derived from `tuple`,
-    //  we can just call `tp_new` on them
+    // since `namedtuple`-s are immutable and derived from `tuple`, we can
+    //  just call `tp_new` on them
     // XXX fix this if the namedtuple's implementation changes
     // https://docs.python.org/3/c-api/typeobj.html#c.PyTypeObject.tp_new
     PyObject *namedtuple = Py_TYPE(main)->tp_new(Py_TYPE(main), output, NULL);
@@ -315,7 +274,8 @@ static PyObject* _apply_list(
     const bool safe,
     const bool star,
     PyObject *kwargs,
-    PyObject *finalizer)
+    PyObject *finalizer,
+    const bool strict)
 {
     Py_ssize_t len = PyTuple_GET_SIZE(rest);
     PyObject *main_, *item_, *rest_ = PyTuple_New(len);
@@ -342,7 +302,7 @@ static PyObject* _apply_list(
             PyTuple_SET_ITEM(rest_, j, item_);
         }
 
-        result = _apply(callable, main_, rest_, safe, star, kwargs, finalizer);
+        result = _apply(callable, main_, rest_, safe, star, kwargs, finalizer, strict);
         if(result == NULL) {
             Py_DECREF(rest_);
             Py_DECREF(output);
@@ -369,7 +329,8 @@ static PyObject* _apply_mapping(
     const bool safe,
     const bool star,
     PyObject *kwargs,
-    PyObject *finalizer)
+    PyObject *finalizer,
+    const bool strict)
 {
     // XXX it's unlikely that we will ever use this branch, because as docs say
     //  it is impossible to know the type of keys of a mapping at runtime, hence
@@ -412,7 +373,7 @@ static PyObject* _apply_mapping(
 
         Py_DECREF(result);
 
-        result = _apply(callable, main_, rest_, safe, star, kwargs, finalizer);
+        result = _apply(callable, main_, rest_, safe, star, kwargs, finalizer, strict);
         if(result == NULL) break;
 
         PyDict_SetItem(output, key, result);
@@ -441,6 +402,8 @@ static PyObject* _apply_base(
     PyObject *item_, *args = PyTuple_New(1+len);
     if(args == NULL) return NULL;
 
+    // reusing `rest` works in our case since the base call actually reassembles
+    //  the full tuple: (main,) + rest
     Py_INCREF(main);
     PyTuple_SET_ITEM(args, 0, main);
     for(Py_ssize_t j = 0; j < len; j++) {
@@ -468,35 +431,38 @@ PyObject* _apply(
     const bool safe,
     const bool star,
     PyObject *kwargs,
-    PyObject *finalizer)
+    PyObject *finalizer,
+    const bool strict)
 {
     PyObject *result;
 
-    if(PyDict_Check(main)) {
+    if(PyDict_CheckExact(main) || (!strict && PyDict_Check(main))) {
         if(safe)
             if(!_validate_dict(main, rest))
                 return NULL;
 
         if(Py_EnterRecursiveCall("")) return NULL;
-        result = _apply_dict(callable, main, rest, safe, star, kwargs, finalizer);
+        result = _apply_dict(callable, main, rest, safe, star, kwargs, finalizer, strict);
         Py_LeaveRecursiveCall();
 
-    } else if(PyTuple_Check(main)) {
+    } else if(
+        PyTupleNamedTuple_CheckExact(main) || (!strict && PyTuple_Check(main))
+    ) {
         if(safe)
             if(!_validate_tuple(main, rest))
                 return NULL;
 
         if(Py_EnterRecursiveCall("")) return NULL;
-        result = _apply_tuple(callable, main, rest, safe, star, kwargs, finalizer);
+        result = _apply_tuple(callable, main, rest, safe, star, kwargs, finalizer, strict);
         Py_LeaveRecursiveCall();
 
-    } else if(PyList_Check(main)) {
+    } else if(PyList_CheckExact(main) || (!strict && PyList_Check(main))) {
         if(safe)
             if(!_validate_list(main, rest))
                 return NULL;
 
         if(Py_EnterRecursiveCall("")) return NULL;
-        result = _apply_list(callable, main, rest, safe, star, kwargs, finalizer);
+        result = _apply_list(callable, main, rest, safe, star, kwargs, finalizer, strict);
         Py_LeaveRecursiveCall();
 
     } else {
@@ -525,7 +491,9 @@ int parse_apply_args(
     PyObject **rest)
 {
     // docs imply that `PyTuple_GetSlice` does not steal references to the elements
-    //  in the slice, meaning that the objects in it are increfed
+    //  in the slice, meaning that the objects in it are increfed. However, when
+    //  slicing yields the full original tuple, in this case the returned object
+    //  is exactly the original tuple, but with its own refcount incremented.
     PyObject *first = PyTuple_GetSlice(args, 0, 2);
     if (first == NULL)
         return 0;
@@ -541,6 +509,7 @@ int parse_apply_args(
         return 0;
     }
 
+    // always increfs the items in the slice, since it always creates a new tuple
     Py_ssize_t len = PyTuple_GET_SIZE(args);
     *rest = PyTuple_GetSlice(args, 2, len);
 
@@ -550,9 +519,9 @@ int parse_apply_args(
 
 PyObject* apply(PyObject *self, PyObject *args, PyObject *kwargs)
 {
-    // from the url at the top: {API 1.2.1} the call mechanism guarantees
+    // from the URL at the top: {API 1.2.1} the call mechanism guarantees
     //  to hold a reference to every argument for the duration of the call.
-    int safe = 1, star = 1;
+    int safe = 1, star = 1, strict=1;
     PyObject *callable = NULL, *main = NULL, *rest = NULL, *finalizer=NULL;
 
     // handle `apply(fn, main, *rest, ...)`
@@ -564,7 +533,13 @@ PyObject* apply(PyObject *self, PyObject *args, PyObject *kwargs)
     // handle `apply(..., *, _star, _safe, _finalizer, **kwargs)`
     // XXX if the finalizer is not required, the kwarg must be omitted.
     if (kwargs) {
-        static char *kwlist[] = {"_safe", "_star", "_finalizer", NULL};
+        static const char *kwlist[] = {
+            "_safe",
+            "_star",
+            "_finalizer",
+            "_strict",
+            NULL,
+        };
 
         // Pop apply's kwargs from `kwargs` so that it could be passed along to
         //  `_apply_base`.
@@ -586,7 +561,9 @@ PyObject* apply(PyObject *self, PyObject *args, PyObject *kwargs)
         // Thus we hold on to the `finalizer` in case its only ref was
         //  the `kwargs`, which we tinkered with just above.
         int parsed = PyArg_ParseTupleAndKeywords(
-            empty, own, "|$ppO:apply", kwlist, &safe, &star, &finalizer);
+            empty, own, "|$ppOp:apply", (char**) kwlist,
+            &safe, &star, &finalizer, &strict
+        );
 
         Py_DECREF(empty);
         if (!parsed) {
@@ -611,7 +588,7 @@ PyObject* apply(PyObject *self, PyObject *args, PyObject *kwargs)
 
     // make the call, then decref everything we might own
     PyObject *result = _apply(
-        callable, main, rest, safe, star, kwargs, finalizer);
+        callable, main, rest, safe, star, kwargs, finalizer, strict);
 
     Py_XDECREF(finalizer);
     Py_DECREF(rest);
