@@ -2,7 +2,7 @@ int parse_apply_args(PyObject *args, PyObject **callable, PyObject **main, PyObj
 
 PyObject* _apply(PyObject *callable, PyObject *main, PyObject *rest,
                  bool const safe, bool const star, PyObject *kwargs,
-                 PyObject *finalizer);
+                 PyObject *finalizer, const bool strict);
 
 PyObject* apply(PyObject *self, PyObject *args, PyObject *kwargs);
 
